SCRIPT_TITLE = "Pitch Deviation Limiter V0.52"

paramDispNames = {
  "no Limiter", "Limiter 1 (Linear function)", "Limiter 2 (Expon. function)", "Limiter 3 (Sinus function)"
}

paramTypeNames = {
  "noLimiter", "linLimiter", "expLimiter", "sinLimiter"
}

function getClientInfo()
  return {
    name = "ak "..SCRIPT_TITLE,
    author = "akStudio GG",
    category = "akStudio",
    versionNumber = 1,
    minEditorVersion = 65537
  }
end

function main()
 
  local form = {
    title = SCRIPT_TITLE,
    message = "",
    buttons = "OkCancel",
    widgets = {
      {
        name = "label1", type = "TextArea",
        label = "------------------ Time range extension ------------------",
        height = 0,
        default = "dummy",
      },
      {
        name = "timeBefore",
        type = "Slider",
        label = "Time before selected Note(s) [% of a quarter]",
        format = "%1.0f",
        minValue = 0,
        maxValue = 150,
        interval = 5,
        default = 50
      },
      {
        name = "timeAfter",
        type = "Slider",
        label = "Time after selected Note(s) [% of a quarter]",
        format = "%1.0f",
        minValue = 0,
        maxValue = 150,
        interval = 5,
        default = 50
      },
      {
        name = "label2", type = "TextArea",
        label = "--------- Deviation modification parameters ---------",
        height = 0,
        default = "dummy",
      },
      {
        name = "linFactor",
        type = "Slider",
        label = "linear Compression / Expansion [%]",
        format = "%1.0f",
        minValue = 0,
        maxValue = 200,
        interval = 5,
        default = 100
      },
      {
        name = "limitType",
        type = "ComboBox",
        label = "Limiter Type",
        choices = paramDispNames,
        default = 3
      },
     {
        name = "limitValue",
        type = "Slider",
        label = "Pitch deviation limit [cent]",
        format = "%1.0f",
        minValue = 5,
        maxValue = 300,
        interval = 5,
        default = 45
      },
      {
        name = "limitFactor4",
        type = "CheckBox",
        text = "Limiter range: Pitch deviation limit x 4",
        default = false
      },
      {
        name = "label3", type = "TextArea",
        label = "Deviation Inversion for values outside the limits:",
        height = 0,
        default = "dummy",
      },
      {
        name = "invStrength",
        type = "Slider",
        label = "Deviation inversion strength",
        format = "%1.1f",
        minValue = 0.0,
        maxValue = 5,
        interval = 0.1,
        default = 1.5
      },
      {
        name = "simplFactor",
        type = "Slider",
        label = "Simplify (Remove points after processing)",
        format = "%1.0f",
        minValue = 0,
        maxValue = 100,
        interval = 1,
        default = 0
      }
    }
  }

  local results = SV:showCustomDialog(form)
  
  if results.status then
    modifyPitch(results.answers)
  end

  SV:finish()
end

-- Get an array of blick ranges for connected notes in the selection
function getSelectedRanges(tLeft,tRight)
  local selection = SV:getMainEditor():getSelection()
  local selectedNotes = selection:getSelectedNotes()
  if #selectedNotes == 0 then
    SV:showMessageBox("Hinweis","keine Note(n) ausgewählt \nAbbruch")
    return {}
  end
  table.sort(selectedNotes, function(noteA, noteB)
    return noteA:getOnset() < noteB:getOnset()
  end)
  
  local ranges = {}
  local bStart = selectedNotes[1]:getOnset() - tLeft
  local bEnd = selectedNotes[1]:getEnd() + tRight
  for i = 2, #selectedNotes do
    if selectedNotes[i]:getOnset() - tLeft  > bEnd then
      ranges[#ranges + 1] = {bStart, bEnd}
      bStart = selectedNotes[i]:getOnset() - tLeft
    end
    bEnd = selectedNotes[i]:getEnd() +  tRight
  end
  ranges[#ranges + 1] = {bStart, bEnd}
  return ranges
end

function editPoint(pitchPoint,parameters)
  local limitType = parameters.limitType
  local limitVal = parameters.limitValue
  if parameters.limitFactor4 then
    limitVal = 4 * limitVal
  end
  local invCoeff = math.max(parameters.invStrength, 0.01) -- to prevent division/zero

  pitchPoint = pitchPoint * parameters.linFactor / 100 -- Pre-gain

  if limitType == 1 then -- Linear function
    if pitchPoint > limitVal then
      pitchPoint = math.max(limitVal - (pitchPoint - limitVal) * invCoeff, 0)
    end
    if pitchPoint < -limitVal then
      pitchPoint = math.min(-limitVal - (pitchPoint + limitVal) * invCoeff, 0)
    end
    
  elseif limitType == 2 then -- Expon.-Function
    if pitchPoint >= 0 then
      pitchPoint = expCurve(pitchPoint, limitVal, invCoeff)
    else
      pitchPoint = -expCurve(-pitchPoint, limitVal, invCoeff)
    end
   
  elseif limitType == 3 then -- Sinus Function
    if pitchPoint >= 0 then
      pitchPoint = sinCurve(pitchPoint, limitVal, invCoeff)
    else
      pitchPoint = -sinCurve(-pitchPoint, limitVal, invCoeff)
    end
   
  end
  
  pitchPoint = math.max(pitchPoint, -1200)
  pitchPoint = math.min(pitchPoint, 1200)
  return math.floor(pitchPoint)
end

function expCurve(x1, v1, v2)
  local k1 = v1 * 1.85 * v2^0.2 -- empirically determined factor
  local t1 = k1 * 0.8           -- empirically determined factor
  local y1 = k1 * (1 - math.exp(-x1/t1))
  local x2 = y1 - x1
  local k2 = k1
  local t2 = k2 * 0.5 / v2
  local y2 =  k2 * (1 - math.exp(-(x2^2 / t2^2)))

  if x2 < 0 then
    y2 = -y2
  end
  
  return y1 + y2
end

function sinCurve(x1, v1, v2)
  local t1 = v1 / 0.7           -- empirically determined factor
  local t2 = v1 / v2
  
  if x1 < t1 then
    return v1 * math.sin(x1 / t1 * math.pi/2)
  elseif x1 < (t1 + t2) then
    return v1 / 2 * (1 + math.cos((x1 - t1) / t2 * math.pi))
  else 
    return 0
  end
  
end

function modifyPitch(options)
  local scope = SV:getMainEditor():getCurrentGroup()
  local group = scope:getTarget()
  local ranges = getSelectedRanges(options.timeBefore / 100 * SV.QUARTER, options.timeAfter / 100 * SV.QUARTER)

  local pitchCurve = group:getParameter("pitchDelta")
	for i, r in ipairs(ranges) do
	  local points = pitchCurve:getPoints(r[1], r[2])
	  for _, p in ipairs(points) do
      pitchCurve:add(p[1], editPoint(p[2],options))
	  end
    
	  pitchCurve:simplify(r[1], r[2],options.simplFactor / 1000)
	end
end
